$dark: (
  primary-color: #9090c0,
  header-color: #e0a830,
  secondary-color: #d4af37,
  page-color: #2d2d3a,
  text-color: #e8e8f0,
  background-color: #1a1a1a,
  border-color: #4a4a5a,
  highlight-color: #b0a0d8,
  code-background-color: #1e1e28,
  inverse-text-color: #2c1810,
  inverse-background-color: #f5deb3,
  inverse-border-color: #8b4513,
  inverse-primary-color: #8b4513,
  border-radius: 5px,
  body-text-shadow: 2px 2px 2px #1a1a1a,
  header-text-shadow: -1px 2px 3px rgba(0, 0, 0, 0.6),
  table-text-color: #e8e8f0,
);
